Decimal32

Struct Decimal32 

pub struct Decimal32<const PRECISION: u32>(/* private fields */);
A 32-bit lossless number with const-defined decimal precision.

Design

There are different ways to represent a floating point number with wrapping and saturating semantics. This design basically takes the bounds of an i32 and sticks a decimal point somewhere. So the type itself serves primarily for self-documentation and convenience.

IEEE 754 floating point debauchery keeps the lossless range of an f32 in signed 2^24. So an equally valid design could lock the inner value within that range and use that bound for wrapping, saturating, and checked operations. The benefit is that get could return an f32 without loss of precision. The cost is hardware support for arithmetic. Since wrapping arithmetic is the primary motivation for Decimal32, this was not chosen.

A Decimal64 type is not (yet) exposed because at that point, the Decimal crate might be a better fit for your use case.

Implementations

Source

impl<const PRECISION: u32> Decimal32<PRECISION>

pub const ZERO: Self

pub const MAX: Self

The greatest positive value this type can contain.

pub const MIN: Self

The most negative value this type can contain.

pub const MIN_UNIT: Self

The smallest positive value this type can contain.

pub const fn cast(value: f64) -> Self

Constructor that accepts any input. Truncates toward zero when the input has more decimal places than PRECISION. Use Self::checked to detect when precision is lost.

This function takes an f64 to prevent sneaky loss of precision. f32 only has a 24-bit mantissa, so values between 2^24 and and 2^31 require an f64 to be constructed.

use high_roller::decimal::D2;

let num = D2::cast(0.125_f64);
assert_eq!(num.get(), 0.12_f64);

If this f64 situation is annoying for your use case, you can still escape it entirely at compile time.

use high_roller::decimal::D3;

const MY_F32: f32 = D3::cast(0.321_f64).get() as f32;
assert_eq!(MY_F32, 0.321);
pub const fn checked(value: f64) -> Option<Self>

Const constructor that prevents loss of precision from the input value.

Succeeds
use high_roller::decimal::D5;

const GOOD: D5 = D5::checked(-100.12345).unwrap();
Fails
use high_roller::decimal::D9;

const BAD: D9 = D9::checked(-100.)
    .expect("There isn't space in 32 bits for 9 decimal places after -100");
pub const fn get(self) -> f64

Returns the inner value as an f64. This conversion is lossless because f64’s 53-bit mantissa can represent every i32 value exactly.

For f32 output use f32::try_from, which returns Err(Lossy) when the inner value exceeds f32’s 24-bit mantissa.

use high_roller::decimal::D1;

assert_eq!(D1::cast(1.0_f64).get(), 1.0_f64);