Module decimal

Module decimal 

Structs

Decimal32
Decimal32

Enums

DecimalErr
Enumerates the possible errors Decimal operations may return.

Type Aliases

D1
A Decimal32 type with one significant figure after the decimal point.
D2
A Decimal32 type with two significant figures after the decimal point.
D3
A Decimal32 type with three significant figures after the decimal point.
D4
A Decimal32 type with four significant figures after the decimal point.
D5
A Decimal32 type with five significant figures after the decimal point.
D6
A Decimal32 type with six significant figures after the decimal point.
D7
A Decimal32 type with seven significant figures after the decimal point.
D8
A Decimal32 type with eight significant figures after the decimal point.
D9
A Decimal32 type with nine significant figures after the decimal point.