RollingSum

Struct RollingSum 

pub struct RollingSum<T, const WINDOW: usize> { /* private fields */ }

Implementations

Source

impl<T, const WINDOW: usize> RollingSum<T, WINDOW>
where T: Default,

pub const fn new(init: T, zero: T) -> Self

Source

impl<T, const WINDOW: usize> RollingSum<T, WINDOW>

pub fn add(&mut self, val: T)

Adds T to the rolling sum, displacing the oldest member if the window is full to capacity.

If adding T causes numerical overflow, subsequent calls to total will return None until window expirations cause underflow commensurate to the overflow.

Panics

This function panics if the usize variable tracking the number of times the sum has overflowed itself overflows. A window should be sized such that this never occurs.

pub fn total(&self) -> Option<&T>

Returns the accumulated total of all added values that fit within the rolling window’s capacity.

Returns None if the window has overflowed. In that case, it will return to Some(..) when the last element causing overflow is pushed out.