pub struct RollingSum<T, const WINDOW: usize> { /* private fields */ }Implementations
Sourceimpl<T, const WINDOW: usize> RollingSum<T, WINDOW>where
T: Default,
impl<T, const WINDOW: usize> RollingSum<T, WINDOW>where
T: Default,
Sourceimpl<T, const WINDOW: usize> RollingSum<T, WINDOW>
impl<T, const WINDOW: usize> RollingSum<T, WINDOW>
Sourcepub fn add(&mut self, val: T)
pub fn add(&mut self, val: T)
Adds T to the rolling sum, displacing the oldest
member if the window is full to capacity.
If adding T causes numerical overflow, subsequent
calls to total will return None until window
expirations cause underflow commensurate to the overflow.
Panics
This function panics if the usize variable tracking the
number of times the sum has overflowed itself overflows.
A window should be sized such that this never occurs.