RollingSum

Struct RollingSum 

pub struct RollingSum<T> { /* private fields */ }

Implementations

Source

impl<T> RollingSum<T>

pub fn new(capacity: NonZeroUsize) -> Self

pub fn add(&mut self, val: T) -> bool

Adds T to the rolling sum, displacing the oldest member if the window is full to capacity.

If adding T causes numerical overflow, subsequent calls to total will fail until the sum returns to an un-overflowed state.

pub fn total(&self) -> Option<&T>

Returns the accumulated total of all added values that fit within the rolling window’s capacity.

Returns None if the window has overflowed. In that case, it will return to Some(..) when the last element causing overflow is pushed out.